{ Oplossing van de opgave 'Rekeningrijden in Mobilia
  Joris van Rantwijk, Apr 2001
  Algoritme: 'greedy' zoekstrategie }

program Mobilia;

const
  maxstad = 100;

var
  nstad, nweg: Integer;
  nbweg, nmbweg: array[1..maxstad] of Integer;
  bweg, mbweg: array[1..maxstad, 1..maxstad] of Integer;

  { Geeft aan of er een tolpoort in de stad gepland is }
  tolpoort: array[1..maxstad] of Boolean;

  { Geeft aan of het wel/niet tolpoort al vast ligt.
    0: nog vrij, 1: ligt vast,
    2: ligt vast als deze ronde slaagt }
  fixed: array[1..maxstad] of Byte;

{ Plaats wel/geen tolpoort in een stad, en werk recursief
  de consequenties uit. }
function PutTolpoort(i: Integer; v: Boolean): Boolean;
var
  j: Integer;
begin
  if fixed[i] > 0 then begin
    PutTolpoort := (tolpoort[i] = v);
    Exit;
  end; {if}

  fixed[i] := 2;
  tolpoort[i] := v;

  if v then begin
    for j := 1 to nmbweg[i] do begin
      if not PutTolpoort(mbweg[i,j], false) then begin
        PutTolpoort := false;
        Exit;
      end; {if}
    end; {for}
  end {if}
  else begin
    for j := 1 to nbweg[i] do begin
      if not PutTolpoort(bweg[i,j], true) then begin
        PutTolpoort := false;
        Exit;
      end; {if}
    end; {for}
  end; {if}

  PutTolpoort := true;
end; {PutTolpoort}

function Solve: Boolean;
var
  i, j: Integer;
begin
  { Begin met alle steden vrij }
  for i := 1 to nstad do
    fixed[i] := 0;

  for i := 1 to nstad do begin
    { Bezoek alle nog vrije steden }
    if fixed[i] = 0 then begin

      { Probeer om geen tolpoort hier te plaatsen }
      if not PutTolpoort(i, false) then begin

        { Mislukt. Geef de gewijzigde steden weer vrij. }
        for j := 1 to nstad do
          if fixed[j] > 1 then
            fixed[j] := 0;

        { Probeer dan om hier wel een tolpoort te plaatsen }
        if not PutTolpoort(i, true) then begin
          Solve := false;
          Exit;
        end; {if}
      end; {if}

      { Fixeer de steden die in deze ronde vast kwamen te liggen }
      for j := 1 to nstad do
        if fixed[j] > 0 then
          fixed[j] := 1;

    end; {if}
  end; {for}

  Solve := true;
end; {Solve}

{ Ga na of de gevonden oplossing uniek is.
  Maw, probeer nog een andere oplossing te vinden. }
function UniqSolution: Boolean;
var
  firstsol: array[1..maxstad] of Boolean;
  verschil: Boolean;
  i, j: Integer;
begin
  { Bewaar bestaande oplossing }
  for i := 1 to nstad do
    firstsol[i] := tolpoort[i];

  verschil := False;

  { Begin weer met alle steden vrij }
  for i := 1 to nstad do
    fixed[i] := 0;

  for i := 1 to nstad do begin
    { Bezoek alle nog vrije steden }
    if fixed[i] = 0 then begin

      { Probeer om het tegenstelde te doen van in de vorige oplossing }
      if PutTolpoort(i, not firstsol[i]) then begin
        { Ja! We hebben een verschil }
        verschil := True;
      end {if}
      else begin
        { Mislukt. Geef de gewijzigde steden weer vrij. }
        for j := 1 to nstad do
          if fixed[j] > 1 then
            fixed[j] := 0;

        { Probeer dan maar weer hetzelfde }
        if not PutTolpoort(i, firstsol[i]) then begin
          { Geen oplossing. Dat kan natuurlijk niet. }
          WriteLn(StdErr, 'Whaa!! Tweede keer geen oplossing');
          Halt(1);
        end; {if}
      end; {if}

      { Fixeer de steden die in deze ronde vast kwamen te liggen }
      for j := 1 to nstad do
        if fixed[j] > 0 then
          fixed[j] := 1;

    end; {if}
  end; {for}

  UniqSolution := not verschil;
end; {UniqSolution}

procedure MaakBWeg(i, j: Integer);
begin
  if nbweg[i] > nstad then begin
    WriteLn(StdErr, 'Fout in invoer');
    Halt(1);
  end; {if}
  Inc(nbweg[i]);
  bweg[i, nbweg[i]] := j;
  if nbweg[j] > nstad then begin
    WriteLn(StdErr, 'Fout in invoer');
    Halt(1);
  end; {if}
  Inc(nbweg[j]);
  bweg[j, nbweg[j]] := i;
end; {MaakBWeg}

procedure MaakMBWeg(i, j: Integer);
begin
  if nmbweg[i] > nstad then begin
    WriteLn(StdErr, 'Fout in invoer');
    Halt(1);
  end; {if}
  Inc(nmbweg[i]);
  mbweg[i, nmbweg[i]] := j;
  if nmbweg[j] > nstad then begin
    WriteLn(StdErr, 'Fout in invoer');
    Halt(1);
  end; {if}
  Inc(nmbweg[j]);
  mbweg[j, nmbweg[j]] := i;
end; {MaakMBWeg}

procedure PrintPlaatsing;
var
  eerstestad: Boolean;
  i: Integer;
begin
  eerstestad := True;
  for i := 1 to nstad do begin
    if tolpoort[i] then begin
      if not eerstestad then
        Write(' ');
      eerstestad := False;
      Write(i);
    end; {if}
  end; {for}
  WriteLn;
end; {PrintPlaatsing}

var
  nrun, run: Integer;
  i, j, k: Integer;
  s: string;

begin
  ReadLn(nrun);

  for run := 1 to nrun do begin

    ReadLn(nstad, nweg);
    if (nstad < 2) or (nstad > maxstad) or (nweg < 1) then begin
      WriteLn(StdErr, 'Fout in invoer');
      Halt(1);
    end; {if}

    for i := 1 to nstad do begin
      nbweg[i] := 0;
      nmbweg[i] := 0;
    end; {for}

    for k := 1 to nweg do begin
      Read(i);
      Read(j);
      ReadLn(s);
      if (i < 1) or (j < 1) or (i > nstad) or (j > nstad) then begin
        WriteLn(StdErr, 'Fout in invoer');
        Halt(1);
      end; {if}
      while (Length(S) > 0) and (S[1] = ' ') do
        Delete(S, 1, 1);
      if s = 'B' then
        MaakBWeg(i, j)
      else if s = 'MB' then
        MaakMBWeg(i, j)
      else begin
        WriteLn(StdErr, 'Fout in invoer (', s, ')');
        Halt(1);
      end; {if}
    end; {for}

    if Solve then begin
      PrintPlaatsing;
      if not UniqSolution then begin
        WriteLn(StdErr, '!! Geen unieke oplossing !!');
        PrintPlaatsing;
      end; {if}
    end {if}
    else begin
      WriteLn('NO');
    end; {if}
      
  end; {for}

end. {Mobilia}
