%
% probleem: mobilia
%
% programma programmeerkampioenschappen 2001
%
%

%
% findsolution(BLst,MBLst,GateLst)
%
% find solution for the mobilia instance with the list BLst of endpoints
% (X,Y) of B-ways and MBLst as the list of endpoints of MB-ways.
% GateLst is the list of places where a gate can be placed satisfying the constraints
% posed by B and MB roads
%

findsolution(BLst,MBLst,GateLst) :-
				find_gates(BLst,MBLst,[],[],GateLst).

% ---------------------------------------------------------------------
% find_gates(B,MB,Initial,Forbidden,GateLst) :
%
%	B         : lijst eindpunten belangrijke wegen,
% MB        : lijst eindpunten minder belangrijke wegen
% Initial   : initiele lijst van tolpoort plaatsen  [start lege lijst]
% Forbidden : list of forbidden places (initially empty)
% GateLst   : final list of places
% ---------------------------------------------------------------------

find_gates([],MB,GateLst,Forbidden,GateLst):-nl.

find_gates([(X,Y)|B],MB,GateLst,Forbidden,Output) :-
   disjoint(Forbidden,Output),
    % select X or Y as a candidate place to be included in the GateLst
	  propagate([X],B,MB,Forbidden,Lst,B1,MB1,NewForbidden),
    % we have been successful in propagating X and found that Lst
    % is the set of places to be added
    union(Lst,GateLst,NwGateLst),
    % add the list of candidates placed
    find_gates(B1,MB1,NwGateLst,NewForbidden,Output).
    % try to find a next candidate

   find_gates([(X,Y)|B],MB,GateLst,Forbidden,Output) :-
   propagate([Y],B,MB,Forbidden,LstY,B1,MB1,NewForbidden),
% select the alternative Y if the first place fails
   propagate([Y],B,MB,Forbidden,LstY,B1,MB1,NewForbidden),
   union(LstY,GateLst,NwGateLst),
   find_gates(B1,MB1,NwGateLst,NewForbidden,Output).

propagate([],B,MB,F,[],B,MB,F).

propagate(L,B,MB,F,NwL,B1,MB1,NwF) :-
    disjoint(L,F),!,
      % no selection of forbidden places is allowed
      % propagate the selection through the B-ways
    propB(L,B,B0),!,
      % propagate the selection through MB-ways,
    propMB(L,MB,Neg,MB01),!,
      % propagate the choices through the MB ways,
      % every place X having exactly one endpoint in common with an end
      % point X of an MB way (X,Y) forces Y to be included in a list Neg
      % of places that cannot occur as places for gates
    check(Neg,B),!, % check the compatibility,
    %write('Neg, B = '), write(Neg),write(B), nl,
    union(Neg,F,NwF1),!,
      % add this list to the list of Negatives
    propNeg(NwF1,B0,Lst,B01),!,
    check(Lst,MB),!,
    %write('Lst, MB = '), write(Lst),write(MB), nl,
      % process the Negatives through the B-ways,
      % this in turn results in a number of places that has to be selected
      % as gates
    disjoint(NwF1,Lst),!,
    propagate(Lst,B01,MB01,NwF1,NwLst,B1,MB1,NwF),!,
      % propagate all candidates left
    union(L,NwLst,NwL),!,
    disjoint(NwF,NwL),!,
    check(NwF,B),!, check(NwL,MB),!.
    %write('NwF, B = '), write(Lst),write(B), nl,
    %write('NewL, MB = '), write(NewL),write(MB), nl.

check([],B) :-!.
check(L,[]) :- !.
check(L,[(X,Y)|B]) :- not((member(X,L),member(Y,L))),!,check(L,B).


      % append all candidates that could be included

% propB(L,B,B0) : propagate a list of places through B and MB
% for every candidate place in L that also occurs as endpoint of
% a B-way, this b-way is not considered anymore, the remaining ways are collected in B0

propB(L,[],[]).
propB(L,[(X,Y)|B],B1) :- (member(X,L);member(Y,L)),!, propB(L,B,B1).
propB(L,[Z|B],[Z|B1]) :- propB(L,B,B1).

%  propMB(X,MB,L,MB1) : L is the list of literals that is not allowed to occur
%   in the Gatelst..
%   MB1 is the list of ways having no endpoints in L

propMB(Lst,[],[],[]).
propMB(Lst,[(X,Y)|MB],[Y|L],MB1):-
     member(X,Lst),not(member(Y,Lst)),!,propMB(Lst,MB,L,MB1).
propMB(Lst,[(X,Y)|MB],[X|L],MB1) :-
      member(Y,Lst),not(member(X,Lst)),!,propMB(Lst,MB,L,MB1).
propMB(Lst,[Z|MB],L,[Z|MB1]) :- propMB(Lst,MB,L,MB1).

% propNeg(N,B,Lst,B01) : we use the list of forbidden places to select
% places that have to occur in order to satisfy the constraints given by the B-ways

propNeg(N,[],[],[]).
propNeg(N,[(X,Y)|B],NwLst,B1) :- member(X,N), not(member(Y,N)),!,
          propNeg(N,B,Lst,B1), add(Y,Lst,NwLst).
propNeg(N,[(X,Y)|B],NwLst,B1) :- member(Y,N), not(member(X,N)),!,
          propNeg(N,B,Lst,B1),!, add(X,Lst,NwLst).
propNeg(N,[(X,Y)|B],Lst,[(X,Y)|B1]) :- not(member(X,N)), 
not(member(Y,N)),!,propNeg(N,B,Lst,B1).

%
% primitive predicates
%
member(X,[X|L]) :- !.
member(X,[Y|L]) :-member(X,L).

union([],L,L) :- !.
union(L,[],L) :- !.
union([X|L],M,N) :- member(X,M),!,union(L,M,N).
union([X|L],M,[X|N]) :- union(L,M,N).

% Joris: SWI-Prolog gave error messages on the definition of append ...
% append([],L,L).
% append([X|L],M,[X|N]) :- append(L,M,N).

disjoint([],L).
disjoint(L,[]).
disjoint([X|L],M) :- not(member(X,M)), !,disjoint(L,M).

add(X,L,L) :- member(X,L),!.
add(X,L,[X|L]).
