/*
 *  Oplossing voor de opgave "Pacman".
 *  Joris van Rantwijk, Mei 2001.
 *  Algoritme: terugrekenen vanaf het eindspel.
 */

#include <stdio.h>
#include <assert.h>

#undef DEBUG

#define MAXWIDTH	20
#define MAXHEIGHT	20

int w, h;
int pacx, pacy, spookx, spooky;

char doolhof[MAXWIDTH][MAXHEIGHT];	/* 'X'=muur, 'U'=uitgang, '.'=veld */

int pos[2][MAXWIDTH][MAXHEIGHT][MAXWIDTH][MAXHEIGHT];
/* pos[turn][pacx][pacy][spookx][spooky]
   turn = 0: pacman, 1: spook
   pos = 0: onbepaald, -1: spook wint, n: pacman wint in n beurten */


int valid_position(int px, int py, int sx, int sy)
{
	return ( px >= 0 && py >= 0 && px < w && py < h &&
	         doolhof[px][py] != 'X' &&
	         sx >= 0 && sy >= 0 && sx < w && sy < h &&
	         doolhof[sx][sy] == '.' );
}


/*
 *  Is deze zet beter voor Pacman dan wat we hadden ?
 */
int pacman_beste_zet(int px, int py, int sx, int sy, int b)
{
	int z;
	if (!valid_position(px, py, sx, sy))
		return b;
	z = pos[1][px][py][sx][sy];
	return ( b == -1 || (b == 0 && z > 0) || (z > 0 && z < b) ) ? z : b;
}


/*
 *  Is deze zet beter voor het spook dan wat we hadden ?
 */
int spook_beste_zet(int px, int py, int sx, int sy, int b)
{
	int z;
	if (!valid_position(px, py, sx, sy))
		return b;
	z = pos[0][px][py][sx][sy];
	return ( b == -1 || (b == 0 && z > 0) || (z > 0 && z < b) ) ? b : z;
}


void updated(int turn, int px, int py, int sx, int sy);

/*
 *  Bereken de waarde van deze situatie opnieuw.
 */
void fixup(int turn, int px, int py, int sx, int sy)
{
	int a, b;

	if ( !valid_position(px, py, sx, sy) )
		return;

	a = pos[turn][px][py][sx][sy];		/* Oude waarde */

	b = pos[!turn][px][py][sx][sy];		/* Beurt overslaan */
	if (turn == 0) {
		b = pacman_beste_zet(px+1, py, sx, sy, b);	/* Rechts */
		b = pacman_beste_zet(px-1, py, sx, sy, b);	/* Links */
		b = pacman_beste_zet(px, py+1, sx, sy, b);	/* Omlaag */
		b = pacman_beste_zet(px, py-1, sx, sy, b);	/* Omhoog */
		if ( a == 0 || (b > 0 && b < a) )
			a = b;
	} else {
		b = spook_beste_zet(px, py, sx+1, sy, b);	/* Rechts */
		b = spook_beste_zet(px, py, sx-1, sy, b);	/* Links */
		b = spook_beste_zet(px, py, sx, sy+1, b);	/* Omlaag */
		b = spook_beste_zet(px, py, sx, sy-1, b);	/* Omhoog */
		if (b > 0)
			b++;
		if ( a == 0 || (b > 0 && b < a) )
			a = b;
	}

	if (a != pos[turn][px][py][sx][sy]) {
		pos[turn][px][py][sx][sy] = a;
		updated(turn, px, py, sx, sy);
	}
}


/*
 *  De aangegeven positie heeft zojuist een waarde gekregen,
 *  ga na of op basis hiervan nog andere situaties berekend kunnen worden.
 */
void updated(int turn, int px, int py, int sx, int sy)
{
	if (turn) {
		fixup(0, px, py, sx, sy);
		fixup(0, px+1, py, sx, sy);
		fixup(0, px-1, py, sx, sy);
		fixup(0, px, py+1, sx, sy);
		fixup(0, px, py-1, sx, sy);
	} else {
		fixup(1, px, py, sx, sy);
		fixup(1, px, py, sx+1, sy);
		fixup(1, px, py, sx-1, sy);
		fixup(1, px, py, sx, sy+1);
		fixup(1, px, py, sx, sy-1);
	}
}


int zoek_route(void)
{
	int px, py, sx, sy, k;

	/* Zet alles op onbepaald */
	for (px = 0; px < w; px++)
	    for (py = 0; py < h; py++)
		for (sx = 0; sx < w; sx++)
		    for (sy = 0; sy < h; sy++) {
			pos[0][px][py][sx][sy] = 0;
			pos[1][px][py][sx][sy] = 0;
		    }

	/* Als Pacman is opgegeten, wint het spook */
	for (px = 0; px < w; px++)
	    for (py = 0; py < h; py++)
	        if (valid_position(px, py, px, py)) {
	        	pos[0][px][py][px][py] = -1;
	        	pos[1][px][py][px][py] = -1;
	        	updated(0, px, py, px, py);
	        	updated(1, px, py, px, py);
		}

	/* Als Pacman de uitgang bereikt heeft, wint Pacman */
	for (px = 0; px < w; px++)
	    for (py = 0; py < h; py++)
		if (doolhof[px][py] == 'U')
		    for (sx = 0; sx < w; sx++)
			for (sy = 0; sy < h; sy++)
			    if (valid_position(px, py, sx, sy)) {
			        pos[1][px][py][sx][sy] = 1;
			        updated(1, px, py, sx, sy);
			    }

	k = pos[0][pacx][pacy][spookx][spooky];

#ifdef DEBUG
	/* Print een trace van deze partij */
	if (k > 0) {
		int a = k;
		px = pacx; py = pacy; sx = spookx; sy = spooky;
		printf("%d %d %d %d\n", px, py, sx, sy);
		while (1) {
			if (valid_position(px, py, sx, sy) && pos[1][px][py][sx][sy] == a) ;
			else if (valid_position(px+1, py, sx, sy) && pos[1][px+1][py][sx][sy] == a) px++;
			else if (valid_position(px-1, py, sx, sy) && pos[1][px-1][py][sx][sy] == a) px--;
			else if (valid_position(px, py+1, sx, sy) && pos[1][px][py+1][sx][sy] == a) py++;
			else if (valid_position(px, py-1, sx, sy) && pos[1][px][py-1][sx][sy] == a) py--;
			else assert(0);
			printf("%d %d %d %d\n", px, py, sx, sy);
			a--;
			if (a == 0) break;
			if (valid_position(px, py, sx, sy) && pos[0][px][py][sx][sy] == a) ;
			else if (valid_position(px, py, sx+1, sy) && pos[0][px][py][sx+1][sy] == a) sx++;
			else if (valid_position(px, py, sx-1, sy) && pos[0][px][py][sx-1][sy] == a) sx--;
			else if (valid_position(px, py, sx, sy+1) && pos[0][px][py][sx][sy+1] == a) sy++;
			else if (valid_position(px, py, sx, sy-1) && pos[0][px][py][sx][sy-1] == a) sy--;
			else assert(0);
			printf("%d %d %d %d\n", px, py, sx, sy);
		}
	}
#endif

	return (k > 0) ? k : -1;
}
	
int main(void)
{
	int n, i, j;


	scanf("%d\n", &n);

	while ((n--) > 0) {

		scanf("%d %d\n", &w, &h);
		assert(w > 0 && w <= MAXWIDTH);
		assert(h > 0 && h <= MAXHEIGHT);

		pacx = -1; spookx = -1;

		for (j = 0; j < h; j++) {
			char line[MAXWIDTH+3];
			fgets(line, sizeof(line), stdin);
			for (i = 0; i < w; i++) {
				doolhof[i][j] = line[i];
				switch (doolhof[i][j]) {
				  case 'P':
					assert(pacx == -1);
					pacx = i; pacy = j;
					doolhof[i][j] = '.';
					break;
				  case 'S':
					assert(spookx == -1);
					spookx = i; spooky = j;
					doolhof[i][j] = '.';
					break;
				  case 'X':
				  case 'U':
				  case '.':
					break;
				  default:
					fprintf(stderr, "Ongeldig karakter '%c' !!\n", doolhof[i][j]);
					exit(1);
				}
			}
		}

		assert(pacx != -1);
		assert(spookx != -1);

		i = zoek_route();
		if (i < 0)
			printf("Geen veilige route\n");
		else
			printf("Pacman bereikt de uitgang in beurt %d\n", i);
	}

	return 0;
}

/* end */
